/*
Copyright (c) 2014 Carnegie Mellon University.
All Rights Reserved.
Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following acknowledgments and disclaimers.
2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following acknowledgments and disclaimers in the documentation and/or other materials provided with the distribution.
3. Products derived from this software may not include Carnegie Mellon University, "SEI and/or Software Engineering Institute" in the name of such derived product, nor shall Carnegie Mellon University, "SEI and/or Software Engineering Institute" be used to endorse or promote products derived from this software without prior written permission. For written permission, please contact permission@sei.cmu.edu.
ACKNOWLEDMENTS AND DISCLAIMERS:
Copyright 2014 Carnegie Mellon University

This material is based upon work funded and supported by the Department of Defense under Contract No. FA8721-05-C-0003 with Carnegie Mellon University for the operation of the Software Engineering Institute, a federally funded research and development center.

Any opinions, findings and conclusions or recommendations expressed in this material are those of the author(s) and do not necessarily reflect the views of the United States Department of Defense.

NO WARRANTY. THIS CARNEGIE MELLON UNIVERSITY AND SOFTWARE ENGINEERING INSTITUTE MATERIAL IS FURNISHED ON AN AS-IS BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER EXPRESSED OR IMPLIED, AS TO ANY MATTER INCLUDING, BUT NOT LIMITED TO, WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY, OR RESULTS OBTAINED FROM USE OF THE MATERIAL. CARNEGIE MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.

This material has been approved for public release and unlimited distribution.

Carnegie Mellon is registered in the U.S. Patent and Trademark Office by Carnegie Mellon University.

DM-0000890

 */

package edu.cmu.sei.ZeroSlackBinPacker;

import java.io.FileWriter;
import java.util.Iterator;
import java.util.Random;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.Vector;

public class Packer {
	
	Vector<Processor> bins = new Vector<Processor>();
	Vector<PTask> taskset = new Vector<PTask>();
	
	TreeMap<PTask, TreeSet<Processor>> binsPerTask;
	TreeSet<PTask> inflatedTaskSet=null;

	TreeMap<PTask, TreeSet<Processor>> plainBinsPerTask;
	TreeSet<PTask> plainInflatedTaskSet=null;

	public Vector<Processor> getProcessors(){
		return bins;
	}
	
	public Vector<PTask> getTaskSet(){
		return taskset;
	}
	
	public void setTaskset(Vector<PTask>taskset){
		this.taskset = taskset;
	}
	
	public void addTask(PTask task){
		taskset.add(task);
		
		// reset the inflated task set
		// and bins per task
		inflatedTaskSet = null;
		binsPerTask = null;
		plainInflatedTaskSet = null;
		plainBinsPerTask = null;
	}
	
	public void addProcessor(Processor p){
		bins.add(p);
		// reset the binsPerTask
		binsPerTask = null;
		plainBinsPerTask = null;
	}

	TreeSet<PTask> czTasks = null;
	public Iterator<PTask> getCzOrderedTasks(){
		if (czTasks == null){
			czTasks = new TreeSet<PTask>(new CriticalitySizeComparator());
			czTasks.addAll(taskset);
		}
		return czTasks.iterator();
	}
	
	TreeSet<PTask> cCoTaskset=null;
	
	public Iterator<PTask> getCritCoOrderedTasks(){
		if (cCoTaskset == null){
			cCoTaskset = new TreeSet<PTask>(new CriticalityOverloadComparator());
			cCoTaskset.addAll(taskset);
		}
		return cCoTaskset.iterator();
	}
	
	public Iterator<PTask> getOrderedTasks(){
		if (inflatedTaskSet == null){
			inflatedTaskSet = new TreeSet<PTask>(new BlockingInflatedComparator(taskset));
			inflatedTaskSet.addAll(taskset);
		}
		return inflatedTaskSet.iterator();
	}
	
	public Iterator<PTask> getPlainOrderedTasks(){
		if (plainInflatedTaskSet == null){
			plainInflatedTaskSet = new TreeSet<PTask>(new OverloadedTaskComparator());
			plainInflatedTaskSet.addAll(taskset);
		}
		return plainInflatedTaskSet.iterator();
	}

	public Iterator<PTask> getNormalPlainOrderedTasks(){
		if (plainInflatedTaskSet == null){
			plainInflatedTaskSet = new TreeSet<PTask>(new OverloadedTaskComparator());
			plainInflatedTaskSet.addAll(taskset);
		}
		return plainInflatedTaskSet.iterator();
	}
	
	public Iterator<Processor> getCrtiCoIncOrderedBinList(){
		TreeSet<Processor> binlist = new TreeSet<Processor>(new IncreasingCriticalityOverloadFullnessComparator());
		binlist.addAll(bins);
		return binlist.iterator();
	}

	public Iterator<Processor> getCrtiCoOrderedBinList(){
		TreeSet<Processor> binlist = new TreeSet<Processor>(new CriticalityOverloadFullnessComparator());
		binlist.addAll(bins);
		return binlist.iterator();
	}

	public Iterator<Processor> getCrtiCOrderedBinList(){
		TreeSet<Processor> binlist = new TreeSet<Processor>(new CriticalitySizeFullnessComparator());
		binlist.addAll(bins);
		return binlist.iterator();
	}

	public Iterator<Processor> getCrtiCIncOrderedBinList(){
		TreeSet<Processor> binlist = new TreeSet<Processor>(new IncreasingCriticalitySizeFullnessComparator());
		binlist.addAll(bins);
		return binlist.iterator();
	}

	public Iterator<Processor> getCzOrderedBinListForTask(PTask task){
		if (binsPerTask == null){
			binsPerTask = new TreeMap<PTask,TreeSet<Processor>>(new TaskComparator());
		}
		
		TreeSet<Processor> binlist = binsPerTask.get(task);
		if (binlist == null){
			binlist = new TreeSet<Processor>(new CriticalitySizeFullnessComparator());
			binlist.addAll(bins);
			binsPerTask.put(task,binlist);
		}
		
		return binlist.iterator();
	}

	public Iterator<Processor> getOrderedBinListForTask(PTask task){
		if (binsPerTask == null){
			binsPerTask = new TreeMap<PTask,TreeSet<Processor>>(new TaskComparator());
		}
		
		TreeSet<Processor> binlist = binsPerTask.get(task);
		if (binlist == null){
			binlist = new TreeSet<Processor>(new BlockingInflatedFullnessComparator(task));
			binlist.addAll(bins);
			binsPerTask.put(task,binlist);
		}
		
		return binlist.iterator();
	}

	public Iterator<Processor> getPlainOrderedBinListForTask(PTask task){
		if (plainBinsPerTask == null){
			plainBinsPerTask = new TreeMap<PTask,TreeSet<Processor>>(new TaskComparator());
		}
		
		TreeSet<Processor> binlist = plainBinsPerTask.get(task);
		if (binlist == null){
			binlist = new TreeSet<Processor>(new FullnessProcessorComparator());
			binlist.addAll(bins);
			plainBinsPerTask.put(task,binlist);
		}
		
		return binlist.iterator();
	}

	public Iterator<Processor> getNormalPlainOrderedBinListForTask(PTask task){
		if (plainBinsPerTask == null){
			plainBinsPerTask = new TreeMap<PTask,TreeSet<Processor>>(new TaskComparator());
		}
		
		TreeSet<Processor> binlist = plainBinsPerTask.get(task);
		if (binlist == null){
			binlist = new TreeSet<Processor>(new NormalFullnessProcessorComparator());
			binlist.addAll(bins);
			plainBinsPerTask.put(task,binlist);
		}
		
		return binlist.iterator();
	}

	public void czPack(){
		double minimumUtilization = 0.0;
		for (Iterator<PTask>iter = getCzOrderedTasks();iter.hasNext();){
			PTask t = iter.next();
			minimumUtilization += ((double)t.getNormalC()) / ((double)t.getPeriod());
		}
		
		long minNumProcs =  (long) Math.ceil(minimumUtilization);

		// add minimum number of processors
		for (long i=0;i<minNumProcs; i++)
			addProcessor(new Processor("",taskset));
		
		for (Iterator<PTask>iter = getCzOrderedTasks();iter.hasNext();){
			PTask t = iter.next();
			boolean deployed = false;
			do{
				for (Iterator<Processor> piter = getCzOrderedBinListForTask(t); piter.hasNext();){
					Processor p = piter.next();
					deployed = p.addIfSchedulable(t);
					if (deployed){
						System.out.println("zsPack:Deployed: t("+t.getPeriod()+") to P("+p.getUniqueID()+"): NormalFullness("+p.normalFullnessLevel+")");						
						break;
					}
				}
				if (!deployed){
					// add new processor
					addProcessor(new Processor("",taskset));
				}
			} while (!deployed);
		}
	}

	public void zsPack(){
		double minimumUtilization = 0.0;
		for (Iterator<PTask>iter = getOrderedTasks();iter.hasNext();){
			PTask t = iter.next();
			minimumUtilization += ((double)t.getNormalC()) / ((double)t.getPeriod());
		}
		
		long minNumProcs =  (long)Math.ceil(minimumUtilization);

		// add minimum number of processors
		for (long i=0;i<minNumProcs; i++)
			addProcessor(new Processor("",taskset));
		
		for (Iterator<PTask>iter = getOrderedTasks();iter.hasNext();){
			PTask t = iter.next();
			boolean deployed = false;
			do{
				for (Iterator<Processor> piter = getOrderedBinListForTask(t); piter.hasNext();){
					Processor p = piter.next();
					deployed = p.addIfSchedulable(t);
					if (deployed)
						break;
				}
				if (!deployed){
					// add new processor
					addProcessor(new Processor("",taskset));
				}
			} while (!deployed);
		}
	}

	public void sswzsPack(){
		double minimumUtilization = 0.0;
		for (Iterator<PTask>iter = getCritCoOrderedTasks();iter.hasNext();){
			PTask t = iter.next();
			minimumUtilization += ((double)t.getNormalC()) / ((double)t.getPeriod());
		}
		
		long minNumProcs =  (long)Math.ceil(minimumUtilization);

		// add minimum number of processors
		for (long i=0;i<minNumProcs; i++)
			addProcessor(new DualModeProcessor("",taskset));
		
		// Stage one
		TreeSet<PTask> rest = new TreeSet<PTask>(new DecreasingCriticalityOverloadComparator());
		for (Iterator<PTask>iter = getCritCoOrderedTasks();iter.hasNext();){
			PTask t = iter.next();
			boolean deployed = false;
			for (Iterator<Processor> piter = getCrtiCoIncOrderedBinList(); piter.hasNext();){
				DualModeProcessor p = (DualModeProcessor) piter.next();
				deployed = p.addIfSchedulable(t);
				if (deployed)
					break;
			}
			if (!deployed){
				rest.add(t);
			}
		}

		// add the rest of tasks
		for (Iterator<PTask>iter = rest.iterator();iter.hasNext();){
			PTask t = iter.next();
			boolean deployed = false;
			do{
				for (Iterator<Processor> piter = getCrtiCIncOrderedBinList(); piter.hasNext();){
					DualModeProcessor p = (DualModeProcessor) piter.next();
					p.setCompression(true);
					deployed = p.addIfSchedulable(t);
					if (deployed)
						break;
				}
				if (!deployed){
					// add new processor
					DualModeProcessor dmp =new DualModeProcessor("",taskset);
					dmp.setCompression(true);
					addProcessor(dmp);
				}
			} while (!deployed);
		}		
		// ensure that the ZSRM is run on all processors
		// so we will have an accurate Crm/Cc for each process
		for (Processor p:getProcessors()){
			DualModeProcessor dp = (DualModeProcessor) p;
			dp.setCompression(true);
			dp.isSchedulable();
		}
	}

	public void swzsPack(){
		double minimumUtilization = 0.0;
		for (Iterator<PTask>iter = getCritCoOrderedTasks();iter.hasNext();){
			PTask t = iter.next();
			minimumUtilization += ((double)t.getNormalC()) / ((double)t.getPeriod());
		}
		
		long minNumProcs =  (long)Math.ceil(minimumUtilization);

		// add minimum number of processors
		for (long i=0;i<minNumProcs; i++)
			addProcessor(new DualModeProcessor("",taskset));
		
		// Stage one
		TreeSet<PTask> rest = new TreeSet<PTask>(new CriticalityOverloadComparator());
		for (Iterator<PTask>iter = getCritCoOrderedTasks();iter.hasNext();){
			PTask t = iter.next();
			boolean deployed = false;
			for (Iterator<Processor> piter = getCrtiCoOrderedBinList(); piter.hasNext();){
				DualModeProcessor p = (DualModeProcessor) piter.next();
				deployed = p.addIfSchedulable(t);
				if (deployed)
					break;
			}
			if (!deployed){
				rest.add(t);
			}
		}

		// add the rest of tasks
		for (Iterator<PTask>iter = rest.iterator();iter.hasNext();){
			PTask t = iter.next();
			boolean deployed = false;
			do{
				for (Iterator<Processor> piter = getCrtiCOrderedBinList(); piter.hasNext();){
					DualModeProcessor p = (DualModeProcessor) piter.next();
					p.setCompression(true);
					deployed = p.addIfSchedulable(t);
					if (deployed)
						break;
				}
				if (!deployed){
					// add new processor
					DualModeProcessor dmp =new DualModeProcessor("",taskset);
					dmp.setCompression(true);
					addProcessor(dmp);
				}
			} while (!deployed);
		}
		
		// ensure that the ZSRM is run on all processors
		// so we will have an accurate Crm/Cc for each process
		for (Processor p:getProcessors()){
			DualModeProcessor dp = (DualModeProcessor) p;
			dp.setCompression(true);
			dp.isSchedulable();
		}
	}
	
	public void plainPack(){
		double minimumUtilization = 0.0;
		for (Iterator<PTask>iter = getPlainOrderedTasks();iter.hasNext();){
			PTask t = iter.next();
			minimumUtilization += ((double)t.getNormalC()) / ((double)t.getPeriod());
		}
		
		long minNumProcs =  (long)Math.ceil(minimumUtilization);

		// add minimum number of processors
		for (long i=0;i<minNumProcs; i++)
			addProcessor(new Processor("", taskset));
		
		for (Iterator<PTask>iter = getPlainOrderedTasks();iter.hasNext();){
			PTask t = iter.next();
			boolean deployed = false;
			do{
				for (Iterator<Processor> piter = getPlainOrderedBinListForTask(t); piter.hasNext();){
					Processor p = piter.next();
					deployed = p.addIfSchedulable(t);
					if (deployed){
						System.out.println("plainPack:Deployed: t("+t.getPeriod()+") to P("+p.getUniqueID()+"): NormalFullness("+p.normalFullnessLevel+")");						
						break;
					}
				}
				if (!deployed){
					// add new processor
					addProcessor(new Processor("",taskset));
				}
			} while (!deployed);
		}
	}
	public long getMinProcs()
	{
		double minimumUtilization = 0.0;
		for (Iterator<PTask>iter = getPlainOrderedTasks();iter.hasNext();){
			PTask t = iter.next();
			minimumUtilization += ((double)t.getNormalC()) / ((double)t.getPeriod());
		}
		
		long minNumProcs =  (long)Math.ceil(minimumUtilization);
		return minNumProcs;
	}
	public void edfPlainPack(){
		double minimumUtilization = 0.0;
		for (Iterator<PTask>iter = getPlainOrderedTasks();iter.hasNext();){
			PTask t = iter.next();
			minimumUtilization += ((double)t.getNormalC()) / ((double)t.getPeriod());
		}
		
		long minNumProcs =  (long)Math.ceil(minimumUtilization);

		// add minimum number of processors
		for (long i=0;i<minNumProcs; i++)
			addProcessor(new EDFProcessor("",taskset));
		
		for (Iterator<PTask>iter = getPlainOrderedTasks();iter.hasNext();){
			PTask t = iter.next();
			boolean deployed = false;
			do{
				for (Iterator<Processor> piter = getPlainOrderedBinListForTask(t); piter.hasNext();){
					EDFProcessor p = (EDFProcessor) piter.next();
					deployed = p.addIfSchedulable(t);
					if (deployed)
						break;
				}
				if (!deployed){
					// add new processor
					addProcessor(new EDFProcessor("",taskset));
				}
			} while (!deployed);
		}
	}

	public void edfNormalPlainPack(){
		double minimumUtilization = 0.0;
		for (Iterator<PTask>iter = getNormalPlainOrderedTasks();iter.hasNext();){
			PTask t = iter.next();
			minimumUtilization += ((double)t.getNormalC()) / ((double)t.getPeriod());
		}
		
		long minNumProcs =  (long)Math.ceil(minimumUtilization);

		// add minimum number of processors
		for (long i=0;i<minNumProcs; i++)
			addProcessor(new EDFNormalProcessor("",taskset));
		
		for (Iterator<PTask>iter = getNormalPlainOrderedTasks();iter.hasNext();){
			PTask t = iter.next();
			boolean deployed = false;
			do{
				for (Iterator<Processor> piter = getNormalPlainOrderedBinListForTask(t); piter.hasNext();){
					EDFNormalProcessor p = (EDFNormalProcessor) piter.next();
					deployed = p.addIfSchedulable(t);
					if (deployed)
						break;
				}
				if (!deployed){
					// add new processor
					addProcessor(new EDFNormalProcessor("",taskset));
				}
			} while (!deployed);
		}
	}

	public static void testDataStructures(){
		Packer p = new Packer();
		//String name, long T, long Ca, long Cs, long criticality, long RMPriority
		PTask t0 = new PTask("t0",100,50,10,2,0);
		PTask t1 = new PTask("t1",200,100,20,1,1);
		PTask t2 = new PTask("t2",400,200,40,0,2);
		
		p.addTask(t0);
		p.addTask(t1);
		p.addTask(t2);
		
		Processor p0 = new Processor("po",p.getTaskSet());
		Processor p1 = new Processor("p1",p.getTaskSet());
		Processor p2 = new Processor("p2",p.getTaskSet());
		
		p0.addTask(new PTask("t0",100,50,10,2,0));
		p1.addTask(new PTask("t1",200,100,20,1,1));
		p2.addTask(new PTask("t2",400,200,40,0,2));
		
		p.addProcessor(p0);
		p.addProcessor(p1);
		p.addProcessor(p2);
		
		for (Iterator<PTask>iter = p.getOrderedTasks();iter.hasNext();){
			PTask t = iter.next();
			System.out.println("task("+t.name+"): C("+t.getBlockingInflatedUtilization(p.getTaskSet())+")");
		}
		
		for (Iterator<PTask>iter = p.getOrderedTasks();iter.hasNext();){
			PTask t = iter.next();
			System.out.println("Processor for Task("+t.name+")");
			for (Iterator<Processor> piter = p.getOrderedBinListForTask(t); piter.hasNext();){
				Processor taskProc = piter.next();
				System.out.println("\t Proc("+taskProc.name+"): Fullness("+taskProc.getBlockingInflatedFullnessForTask(t)+")");
			}
			System.out.println(" ");
		}		
	}

	public void reassignRMPriorities(){
		TreeSet<PTask> reordered = new TreeSet<PTask>(new RMOrderComparator());
		reordered.addAll(taskset);
		int i=0;
		for (PTask t:reordered){
			t.RmPriority = i++;
		}
	}
	
	public void generateRandomTaskSet(PTask []taskTypes, int numTasks){
		Random rand = new Random();

		for (int taskIndex=0;taskIndex<numTasks;taskIndex++){
			for (int typeIndex = 0; typeIndex < taskTypes.length; typeIndex++){
				int numPartitions = rand.nextInt()%3;

				taskIndex++;

				double restC = taskTypes[typeIndex].getNormalC();
				double restCo = taskTypes[typeIndex].getOverloadedC();

				int i=0;
				for (i=0; i<numPartitions -1 && restC >0;i++){
					double C=0;
					double Co=0;
					double proportion = rand.nextDouble();
					C = Math.ceil(proportion * restC);
					Co = Math.ceil(proportion* restCo);
					if (C==0){
						C=restC;
						Co=restCo;
					}
					restC = restC - C;
					restCo = restCo - Co;
					PTask t = new PTask("t"+Integer.toString(i),
							taskTypes[typeIndex].getPeriod(),
							(long)Co,
							(long)C,
							taskTypes[typeIndex].criticality,
							taskTypes[typeIndex].RmPriority);
					addTask(t);
				}
				if (restC >0){
					PTask t = new PTask("t"+Integer.toString(i),
							taskTypes[typeIndex].getPeriod(),
							(long)restCo,
							(long)restC,
							taskTypes[typeIndex].criticality,
							taskTypes[typeIndex].RmPriority);
					addTask(t);		
				}
			}
		}
	}
	
	public void generateTaskSet(PTask []taskTypes, int numTasks){
		int typeIndex=0;
		
		for (int i=0;i<numTasks;i++){
			PTask t = new PTask("t"+Integer.toString(i),
								taskTypes[typeIndex].getPeriod(),
								taskTypes[typeIndex].getOverloadedC(),
								taskTypes[typeIndex].getNormalC(),
								taskTypes[typeIndex].criticality,
								taskTypes[typeIndex].RmPriority);
			addTask(t);
			//typeIndex = (typeIndex +1) % taskTypes.length;
			if (i != 0 && (i%(numTasks/taskTypes.length) == 0)){
				typeIndex++;
			}
		}
		
		reassignRMPriorities();
	}
	
	public Vector<PTask> getTasksetClone(){
		Vector<PTask> v = new Vector<PTask>();
		
		for (PTask t:taskset){
			PTask clone = new PTask(t.name, t.getPeriod() , t.getOverloadedC(), t.getNormalC(), t.criticality, t.RmPriority);
			v.add(clone);
		}
		
		return v;
	}

	public double getCriticalityResilience(int numCriticalityLevels){
		double resilience=0.0f;
		long numBins = getProcessors().size();
		
		/*
		 * Karthik: Modified the current version of resilience
		
				long [] digits = new long[numCriticalityLevels];
				
				for (int i=0;i<numCriticalityLevels;i++){
					digits[i]=0;
				}
				
				for (Processor p:getProcessors()){
					long minOKCriticality = p.getHarmonicLowestCriticalityKilledFirst();
				
					for (int i=0; i<=minOKCriticality;i++){
						digits[i]++;
					}
				}
				
				for (int i=0;i<numCriticalityLevels;i++){
					resilience += (long) (digits[i]*Math.pow(numBins, (numCriticalityLevels-1) -i));
				}
		*/
		
		for(long ov=(1L<<(numCriticalityLevels-1))-1; ov>=0; ov>>=1){
			//System.out.println("Considering overload " + ov );
			for(Processor p:getProcessors()){
				resilience += p.getResiliency(ov,numCriticalityLevels);
			}
			if(ov == 0)
				break;
		}
		return resilience;
	}
	
	public static Packer testCzPacker(PTask[] taskTypes, int numTasks){
		Packer p = new Packer();
		//p.generateTaskSet(taskTypes, numTasks);
		
//		p.generateRandomTaskSet(taskTypes, numTasks);

		p.czPack();
		
		System.out.println("CZBP+ZSRM:\t Total Number of Processors: "+p.getProcessors().size());
//		for (Processor proc:p.getProcessors()){
//			System.out.println("Proc("+proc.getUniqueID()+"):");
//			double totalOverloaded=0;
//			for (PTask t:proc.getTaskSet()){
//				System.out.println("\t Task("+t.name+"): Co("+t.getOverloadedC()+"), C("+t.getNormalC()+"), T("+t.getPeriod()+") OverloadedUtilization("+t.getOverloadedUtilization()+")");
//				totalOverloaded += t.getOverloadedUtilization();
//			}
//			System.out.println("\t\t Total Overloaded Utilization: "+totalOverloaded);
//		}
		return p;
	}

	public static Packer testZsPacker(PTask[] taskTypes, int numTasks){
		Packer p = new Packer();
		//p.generateTaskSet(taskTypes, numTasks);
//		p.generateRandomTaskSet(taskTypes, numTasks);

		p.zsPack();
		
		System.out.println("ZSBP+ZSRM:\t Total Number of Processors: "+p.getProcessors().size());
//		for (Processor proc:p.getProcessors()){
//			System.out.println("Proc("+proc.getUniqueID()+"):");
//			double totalOverloaded=0;
//			for (PTask t:proc.getTaskSet()){
//				System.out.println("\t Task("+t.name+"): Co("+t.getOverloadedC()+"), C("+t.getNormalC()+"), T("+t.getPeriod()+") OverloadedUtilization("+t.getOverloadedUtilization()+")");
//				totalOverloaded += t.getOverloadedUtilization();
//			}
//			System.out.println("\t\t Total Overloaded Utilization: "+totalOverloaded);
//		}
		return p;
	}
	
	public static Packer testPlainPacker(PTask[] taskTypes, int numTasks){
		Packer p = new Packer();

		//p.generateTaskSet(taskTypes, numTasks);
		
//		p.generateRandomTaskSet(taskTypes, numTasks);
		p.plainPack();
		
		System.out.println("BP+ZSRM:\t Total Number of Processors: "+p.getProcessors().size());
//		for (Processor proc:p.getProcessors()){
//			System.out.println("Proc("+proc.getUniqueID()+"):");
//			double totalOverloaded=0;
//			for (PTask t:proc.getTaskSet()){
//				System.out.println("\t Task("+t.name+"): Co("+t.getOverloadedC()+"), C("+t.getNormalC()+"), T("+t.getPeriod()+") OverloadedUtilization("+t.getOverloadedUtilization()+")");
//				totalOverloaded += t.getOverloadedUtilization();
//			}
//			System.out.println("\t\t Total Overloaded Utilization: "+totalOverloaded);
//		}
		return p;
	}

	public static Packer testEdfPlainPacker(PTask[] taskTypes, int numTasks){
		Packer p = new Packer();

		//p.generateTaskSet(taskTypes, numTasks);
//		p.generateRandomTaskSet(taskTypes, numTasks);
		p.edfPlainPack();
		
		System.out.println("BP+RM:\t\t Total Number of Processors: "+p.getProcessors().size());
//		for (Processor proc:p.getProcessors()){
//			System.out.println("Proc("+proc.getUniqueID()+"):");
//			double totalOverloaded=0;
//			for (PTask t:proc.getTaskSet()){
//				System.out.println("\t Task("+t.name+"): Co("+t.getOverloadedC()+"), C("+t.getNormalC()+"), T("+t.getPeriod()+") OverloadedUtilization("+t.getOverloadedUtilization()+")");
//				totalOverloaded += t.getOverloadedUtilization();
//			}
//			System.out.println("\t\t Total Overloaded Utilization: "+totalOverloaded);
//		}
		return p;
	}

	public static void testOrder(PTask []taskTypes){
		Vector<PTask> taskset=new Vector<PTask>();
		for (int j=0;j<taskTypes.length;j++){
			PTask t = new PTask("t"+Integer.toString(j),
								taskTypes[j].getPeriod(),
								taskTypes[j].getOverloadedC(),
								taskTypes[j].getNormalC(),
								taskTypes[j].criticality,
								taskTypes[j].RmPriority);
			taskset.add(t);
		}
		
		TreeSet<PTask> tasksByInflated = new TreeSet<PTask>(new BlockingInflatedComparator(taskset));
		for (PTask t:taskset){
			tasksByInflated.add(t);
		}
		TreeSet<PTask> tasksOverloaded = new TreeSet<PTask>(new OverloadedTaskComparator());
		for (PTask t:taskset){
			tasksOverloaded.add(t);
		}
		
		System.out.print("Tasks Inflated:\t\t ");
		String comma="";
		for (PTask t:tasksByInflated){
			System.out.print(comma+t.name);
			comma =" ,";
		}
		System.out.println(" ");

		System.out.print("Tasks Overloaded:\t ");
		comma="";
		for (PTask t:tasksOverloaded){
			System.out.print(comma+t.name);
			comma =" ,";
		}
		System.out.println(" ");
		
	}
	
	public static void writeoutFullyNormalizedResilience(Packer [] packers, FileWriter writer,double normalizedResilience, double w1, double w2){
		try{
			String comma="";
			for (int i=0;i<packers.length;i++){
				writer.write(
						comma+
						Double.toString(
								((w1*((packers[i].getCriticalityResilience(3)/normalizedResilience)) +
								(w2*((double)packers[i].getMinProcs()))/((double)packers[i].getProcessors().size())))
								/(w1+w2)
						)
				);
				comma=",";
			}
			writer.write("\n");
		}catch (Exception e){
			e.printStackTrace();
		}
	}

	public static void printoutFullyNormalizedResilience(Packer [] packers, double normalizedResilience, double w1, double w2){
		try{
			String comma="";
			for (int i=0;i<packers.length;i++){
				System.out.print(
						comma+
						Double.toString(
								((w1*((packers[i].getCriticalityResilience(3)/normalizedResilience)) +
								(w2*((double)packers[i].getMinProcs()/((double)packers[i].getProcessors().size())))))
								/(w1+w2)
						)
				);
				comma=",";
			}
			System.out.print("\n");
		}catch (Exception e){
			e.printStackTrace();
		}
	}

	public static void writeoutNormalizedResilience(Packer [] packers, FileWriter writer,double normalizedResilience){
		try{
			String comma="";
			for (int i=0;i<packers.length;i++){
				writer.write(
						comma+
						Double.toString(
								(packers[i].getCriticalityResilience(3)/normalizedResilience)
						)
				);
				comma=",";
			}
			writer.write("\n");
		}catch (Exception e){
			e.printStackTrace();
		}
	}

	public static void main(String args []){
		
		Packer p = new Packer();
		
		
		int [][] criticalityArrays = {
				{0,1,2},
				{0,2,1},
				{1,0,2},
				{1,2,0},
				{2,0,1},
				{2,1,0}
		};

		int numExperiments = 1;//20;
		int [][]avgNumProcs = new int[7][criticalityArrays.length];
		
		for (int i=0;i<7;i++){
			for (int j=0;j<criticalityArrays.length;j++)
			avgNumProcs[i][j]=0;
		}

		FileWriter NormalizedWriter =null;
		FileWriter NotNormalizedWriter =null;
		try{
			NormalizedWriter = new FileWriter("c:/temp/NormalizedResults.csv");
			NotNormalizedWriter = new FileWriter("c:/temp/NotNormalizedResults.csv");
			NormalizedWriter.write("criticality vector, cobpzsrmp,scobpzsrmp,czbpzsrmp,zsbpzsrmp,bpzsrmp,bprmp,bp(normal)rm\n");
			NotNormalizedWriter.write("criticality vector, cobpzsrmp,scobpzsrmp,czbpzsrmp,zsbpzsrmp,bpzsrmp,bprmp,bp(normal)rm\n");
		} catch (Exception e){
			e.printStackTrace();
		}
		for (int j=0;j<numExperiments;j++){

			for (int i=0;i<criticalityArrays.length;i++){
				PTask [] taskTypes = new PTask[3];



				System.out.println("{"+criticalityArrays[i][0]+","+criticalityArrays[i][1]+","+criticalityArrays[i][2]+"}");
				// task types
				taskTypes[0] = new PTask("t0",1000,100,50,criticalityArrays[i][0],0);
				taskTypes[1] = new PTask("t1",2000,200,100,criticalityArrays[i][1],1);
				taskTypes[2] = new PTask("t2",4000,400,200,criticalityArrays[i][2],2);

//				taskTypes[2] = new PTask("t0",100,30,11,criticalityArrays[i][0],0);
//				taskTypes[1] = new PTask("t1",200,60,22,criticalityArrays[i][1],1);
//				taskTypes[0] = new PTask("t2",400,120,44,criticalityArrays[i][2],2);


				System.out.println("Criticality Vector={"+criticalityArrays[i][0]+","+criticalityArrays[i][1]+","+criticalityArrays[i][2]+"}");

				Packer czbpzsrmp = new Packer();
//				czbpzsrmp.generateRandomTaskSet(taskTypes, 30);
				czbpzsrmp.generateTaskSet(taskTypes, 30);
				Packer zsbpzsrmp = new Packer();
				zsbpzsrmp.taskset = czbpzsrmp.getTasksetClone();
				Packer bpzsrmp = new Packer();
				bpzsrmp.taskset = czbpzsrmp.getTasksetClone();
				Packer bprmp = new Packer();
				bprmp.taskset = czbpzsrmp.getTasksetClone();
				Packer cobpzsrmp = new Packer();
				cobpzsrmp.taskset = czbpzsrmp.getTasksetClone();
				Packer scobpzsrmp = new Packer();
				scobpzsrmp.taskset =  czbpzsrmp.getTasksetClone();
				Packer bpnormalrmp = new Packer();
				bpnormalrmp.taskset = czbpzsrmp.getTasksetClone();
				
				czbpzsrmp.czPack();
				zsbpzsrmp.zsPack();
				bpzsrmp.plainPack();
				bprmp.edfPlainPack();
				cobpzsrmp.swzsPack();
				scobpzsrmp.sswzsPack();
				bpnormalrmp.edfNormalPlainPack();
				
				System.out.println("Number of bins");
				System.out.println("SWZSBP+ZSRM:" +cobpzsrmp.getProcessors().size());
				System.out.println("SSWZSBP+ZSRM:" +scobpzsrmp.getProcessors().size());
				System.out.println("CZBP+ZSRM:" +czbpzsrmp.getProcessors().size());
				System.out.println("ZSBP+ZSRM:" +zsbpzsrmp.getProcessors().size());
				System.out.println("BP+ZSRM:" +bpzsrmp.getProcessors().size());
				System.out.println("BP+RM:" +bprmp.getProcessors().size());
				System.out.println("BP+(normal)RM:" +bpnormalrmp.getProcessors().size());
				
				double normalizedResilience= bprmp.getCriticalityResilience(3);
				System.out.println("\nResilience (normalization value):" + bprmp.getCriticalityResilience(3));
/*				System.out.println("SWZSBP+ZSRM :" +(cobpzsrmp.getCriticalityResilience(3)/normalizedResilience)*((double)cobpzsrmp.getMinNumProcs()/(double)cobpzsrmp.getProcessors().size()));
				System.out.println("SSWZSBP+ZSRM :" +(scobpzsrmp.getCriticalityResilience(3)/normalizedResilience)*((double)scobpzsrmp.getMinNumProcs()/(double)scobpzsrmp.getProcessors().size()));
				System.out.println("CZBP+ZSRM:" +(czbpzsrmp.getCriticalityResilience(3)/normalizedResilience)*((double)czbpzsrmp.getMinNumProcs()/(double)czbpzsrmp.getProcessors().size()));
				System.out.println("ZSBP+ZSRM:" +(zsbpzsrmp.getCriticalityResilience(3)/normalizedResilience)*((double)zsbpzsrmp.getMinNumProcs()/(double)zsbpzsrmp.getProcessors().size()));
				System.out.println("BP+ZSRM:" +(bpzsrmp.getCriticalityResilience(3)/normalizedResilience)*((double)bpzsrmp.getMinNumProcs()/(double)bpzsrmp.getProcessors().size()));
				System.out.println("BP+RM:" + 1.0*((double)bprmp.getMinNumProcs()/(double)bprmp.getProcessors().size()));
*/
				//Need to modify this
				/*System.out.println((cobpzsrmp.getCriticalityResilience(3)/normalizedResilience)*((double)cobpzsrmp.getMinProcs()/((double)cobpzsrmp.getProcessors().size()*2)));
				System.out.println((scobpzsrmp.getCriticalityResilience(3)/normalizedResilience)*((double)scobpzsrmp.getMinProcs()/((double)scobpzsrmp.getProcessors().size()*2)));
				System.out.println((czbpzsrmp.getCriticalityResilience(3)/normalizedResilience)*((double)czbpzsrmp.getMinProcs()/((double)czbpzsrmp.getProcessors().size()*2)));
				System.out.println((zsbpzsrmp.getCriticalityResilience(3)/normalizedResilience)*((double)zsbpzsrmp.getMinProcs()/((double)zsbpzsrmp.getProcessors().size()*2)));
				System.out.println((bpzsrmp.getCriticalityResilience(3)/normalizedResilience)*((double)bpzsrmp.getMinProcs()/((double)bpzsrmp.getProcessors().size()*2)));
				System.out.println(1.0*((double)bprmp.getMinProcs()/((double)bprmp.getProcessors().size()*2)));
				System.out.println(" ");
				*/
				Packer[] packers = {cobpzsrmp,scobpzsrmp,czbpzsrmp,zsbpzsrmp,bpzsrmp,bprmp,bpnormalrmp};
				
				try{
				NotNormalizedWriter.write("{"+criticalityArrays[i][0]+":"+criticalityArrays[i][1]+":"+criticalityArrays[i][2]+"},");
				}catch(Exception e){
					System.err.println(e.getMessage());
				}
				
				writeoutNormalizedResilience(
						packers,
						NotNormalizedWriter,
						normalizedResilience);
				try{
				NormalizedWriter.write("{"+criticalityArrays[i][0]+":"+criticalityArrays[i][1]+":"+criticalityArrays[i][2]+"},");
				}catch(Exception e){
					System.err.println(e.getMessage());
				}

				writeoutFullyNormalizedResilience(
						packers,
						NormalizedWriter,
						normalizedResilience, 0, 1);

				printoutFullyNormalizedResilience(
						packers,
						normalizedResilience, 0, 1);
		
				
//				Packer czbpzsrmp = testCzPacker(taskTypes, 30);
//				Packer zsbpzsrmp = testZsPacker(taskTypes, 30);
//				Packer bpzsrmp = testPlainPacker(taskTypes, 30);
//				Packer bprmp = testEdfPlainPacker(taskTypes, 30);

				avgNumProcs[0][i]+=cobpzsrmp.getProcessors().size();
				avgNumProcs[1][i]+=scobpzsrmp.getProcessors().size();
				avgNumProcs[2][i]+=czbpzsrmp.getProcessors().size();
				avgNumProcs[3][i]+=zsbpzsrmp.getProcessors().size();
				avgNumProcs[4][i]+=bpzsrmp.getProcessors().size();
				avgNumProcs[5][i]+=bprmp.getProcessors().size();
				avgNumProcs[6][i]+=bpnormalrmp.getProcessors().size();

				System.out.println(" ");

//				if (i==1){
//				System.out.println("ZSBP+ZSRM Allocations:");
//				for (Processor proc:zsbpzsrmp.getProcessors()){
//				System.out.print("Processor("+proc.name+")={");
//				String comma="";
//				for (PTask task:proc.getTaskSet()){
//				System.out.print(comma+task.getDescriptor());
//				comma=",";
//				}
//				System.out.println("} ");
//				}
//				System.out.println("BP+ZSRM Allocations:");
//				for (Processor proc:bpzsrmp.getProcessors()){
//				System.out.print("Processor("+proc.name+")={");
//				String comma="";
//				for (PTask task:proc.getTaskSet()){
//				System.out.print(comma+task.getDescriptor());
//				comma=",";
//				}
//				System.out.println("} ");
//				}
//				}
			}
		}
		for (int i=0;i<7;i++){
			for (int j=0;j<criticalityArrays.length;j++)
				avgNumProcs[i][j] /= numExperiments;
		}

		for (int k=0;k<criticalityArrays.length;k++){
			System.out.println("SWZSBP+ZSRM("+k+"):"+avgNumProcs[0][k]);
			System.out.println("SSWZSBP+ZSRM("+k+"):"+avgNumProcs[1][k]);
			System.out.println("CSBP+ZSRM("+k+"):"+avgNumProcs[2][k]);
			System.out.println("ZSBP+ZSRM("+k+"):"+avgNumProcs[3][k]);
			System.out.println("BP+ZSRM("+k+"):"+avgNumProcs[4][k]);
			System.out.println("BP+RM("+k+"):"+avgNumProcs[5][k]);
			System.out.println("BP+(normal)RM("+k+"):"+avgNumProcs[6][k]);
		}
//		PTask [] taskTypes = new PTask[3];

		// These task types highlights the benefits of the
		// new bin packer but the scheduler does not make a
		// difference
//		taskTypes[0] = new PTask("t0",100,50,10,2,0);
//		taskTypes[1] = new PTask("t1",200,100,20,1,1);
//		taskTypes[2] = new PTask("t2",400,200,40,0,2);

		// These tasks types highlights the difference between
		// (zspacker+zsscheduler) and (plainpacker+edfscheduler)
//		taskTypes[0] = new PTask("t0",100,50,10,2,0);
//		taskTypes[1] = new PTask("t1",200,100,20,1,1);
//		taskTypes[2] = new PTask("t2",400,201,40,0,2);

//		testZsPacker(taskTypes, 30);
//		testPlainPacker(taskTypes, 30);
//		testEdfPlainPacker(taskTypes, 30);
		//testDataStructures();
		
		try{
			NotNormalizedWriter.close();
			NormalizedWriter.close();
		} catch (Exception e){
			e.printStackTrace();
		}
	}
}
