/*
Copyright (c) 2014 Carnegie Mellon University.
All Rights Reserved.
Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following acknowledgments and disclaimers.
2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following acknowledgments and disclaimers in the documentation and/or other materials provided with the distribution.
3. Products derived from this software may not include Carnegie Mellon University, "SEI and/or Software Engineering Institute" in the name of such derived product, nor shall Carnegie Mellon University, "SEI and/or Software Engineering Institute" be used to endorse or promote products derived from this software without prior written permission. For written permission, please contact permission@sei.cmu.edu.
ACKNOWLEDMENTS AND DISCLAIMERS:
Copyright 2014 Carnegie Mellon University

This material is based upon work funded and supported by the Department of Defense under Contract No. FA8721-05-C-0003 with Carnegie Mellon University for the operation of the Software Engineering Institute, a federally funded research and development center.

Any opinions, findings and conclusions or recommendations expressed in this material are those of the author(s) and do not necessarily reflect the views of the United States Department of Defense.

NO WARRANTY. THIS CARNEGIE MELLON UNIVERSITY AND SOFTWARE ENGINEERING INSTITUTE MATERIAL IS FURNISHED ON AN AS-IS BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER EXPRESSED OR IMPLIED, AS TO ANY MATTER INCLUDING, BUT NOT LIMITED TO, WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY, OR RESULTS OBTAINED FROM USE OF THE MATERIAL. CARNEGIE MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.

This material has been approved for public release and unlimited distribution.

Carnegie Mellon is registered in the U.S. Patent and Trademark Office by Carnegie Mellon University.

DM-0000890

 */

package edu.cmu.sei.ZeroSlackBinPacker;

import java.util.TreeSet;
import java.util.Vector;

import edu.cmu.sei.ZeroSlackRM.Task;
import edu.cmu.sei.ZeroSlackRM.ZSRMScheduler;

public class Processor {
		
	Vector<PTask> taskset = new Vector<PTask>();
	
	Vector<PTask> globalTaskset = new Vector<PTask>();

	public Vector<PTask> getGlobalTaskset(){
		return globalTaskset;
	}
	
	static long maxUniqueID=0;
	
	double fullnessLevel = 0.0;
	double normalFullnessLevel = 0.0;
	
	public double getNormalFullnessLevel(){
		return normalFullnessLevel;
	}
	
	public double getFullnessLevel(){
		return fullnessLevel;
	}
	
	long uniqueID = maxUniqueID++;
	
	public long getUniqueID(){
		return uniqueID;
	}
	
	public Vector<PTask> getTaskSet(){
		return taskset;
	}
	
	String name;
	
	public String getName(){
		return name;
	}
	
	public Processor(String name,Vector<PTask> gts){
		this.name = name;
		this.globalTaskset = gts;
	}
	
	public void addTask(PTask t){
		taskset.add(t);
		fullnessLevel += t.getOverloadedUtilization();
		normalFullnessLevel += t.getNormalUtilization();
	}
	
	public void removeTask(PTask t){
		taskset.remove(t);
		fullnessLevel -= t.getOverloadedUtilization();
		normalFullnessLevel -= t.getNormalUtilization();
	}
	
	public boolean isSchedulable(){
		ZSRMScheduler scheduler = new ZSRMScheduler();
		
		for (PTask t:taskset){
			scheduler.addTask(t);
		}
		
		scheduler.compressSchedule();
		
		boolean schedulable = true;
		for (Task t:scheduler.getTasksByPriority()){
			schedulable &= (t.ZeroSlackInstant >=0);
		}
		
		return schedulable;
	}

	public boolean addIfSchedulable(PTask t){
		if (t.getNormalC() > t.getOverloadedC()|| t.getPeriod()< t.getOverloadedC()||
				t.getNormalC() <= 0)
			System.out.println("Bad task:C("+t.getNormalC()+") Co("+t.getOverloadedC()+"), T("+t.getPeriod()+")");
		addTask(t);
		if (!isSchedulable()){
			removeTask(t);
			return false;
		}
		return true;
	}
	
	double getBlockingInflatedFullnessForTask(PTask task){
		double fullness=0.0;
		double blocking=0.0;
		
		for (PTask t:taskset){
			fullness += ((double)t.getOverloadedC())/((double)t.getPeriod());
		}
		blocking += task.getTotalBlockingUtilization(taskset);
		//blocking += task.getTotalBlockingUtilization(globalTaskset);
		
		return fullness+blocking;
	}
	
	double getBlockingFullnessForTask(PTask task){
		return task.getTotalBlockingUtilization(taskset);
		//return task.getTotalBlockingUtilization(globalTaskset);
	}
	
	public long getHarmonicLowestCriticalityKilledFirst(){
		
		TreeSet<PTask> tasksByCriticality = new TreeSet<PTask>(new IncreasingCriticalityComparator());		
		tasksByCriticality.addAll(taskset);
		
		// assign highest criticality (zero)
		long lowestCrit = 0;

		for (PTask t:tasksByCriticality){
			double preemption=0;
			
			for (PTask pt:taskset){
				// preemption from lower priority higher or same criticality
				if (pt.criticality <= t.criticality && pt.RmPriority > t.RmPriority){
					preemption += ((double)(pt.getOverloadedC()-pt.getCrm())/(double)pt.getPeriod());
				} else if (pt.RmPriority <= t.RmPriority && pt.criticality> t.criticality){
					preemption += ((double)pt.getCrm()/(double)pt.getPeriod());
				} else if (pt.RmPriority <= t.RmPriority){
					preemption += ((double)pt.getOverloadedC()/(double) pt.getPeriod());
				}				
			}
			
			if (t.getOverloadedUtilization()+preemption <=1.0){
				lowestCrit = t.criticality;
			}
		}
		
		return lowestCrit;
	}
	public double getResiliency(long overload, int numCriticalityLevels){
		double resiliency = 0;
		long[] numTasksPerCriticality = new long[numCriticalityLevels];
		TreeSet<PTask> tasksByCriticality = new TreeSet<PTask>(new IncreasingCriticalityComparator());		
		tasksByCriticality.addAll(taskset);
		
		for(int i=0; i<numCriticalityLevels; i++){
			numTasksPerCriticality[i] = 0;
		}
		
		for (PTask t:globalTaskset){
			if(t.criticality >= numCriticalityLevels){
				System.err.println("ERROR IN ASSUMPTIONS - CRITICALITY LEVEL IS GREATER THAN NUMCRITICALITYLEVELS");
				return 0;
			}
			numTasksPerCriticality[(int)t.criticality] ++;
		}
		for (PTask t:tasksByCriticality){
			double preemption=0;
			
			for (PTask pt:taskset){
				if(t.getUniqueID() == pt.getUniqueID())
					continue;
				// preemption from lower priority higher or same criticality
				if (pt.criticality <= t.criticality && pt.RmPriority > t.RmPriority){
					//check if it is active in our overload condition
					if((overload&(1L<<pt.criticality))>0)
					{
						//Active in our overload scenario
						preemption += ((double)(pt.getOverloadedC()-pt.getCrm())/(double)pt.getPeriod());
					}
				} else if (pt.RmPriority <= t.RmPriority && pt.criticality> t.criticality){
					//higher priority but lower criticality
					preemption += ((double)pt.getCrm()/(double)pt.getPeriod());
				} else if (pt.RmPriority <= t.RmPriority){
					//higher priority and higher criticality
					if((overload&(1L<<pt.criticality))>0)
					{
						preemption += ((double)pt.getOverloadedC()/(double) pt.getPeriod());
					}
					else
					{
						preemption += ((double)pt.getNormalC()/(double)pt.getPeriod());	
					}
				}				
			}
			
			if (t.getOverloadedUtilization()+preemption <=1.0){
				//UNKNOWN EXPLORATION resiliency += ((Math.pow(0.5,(double)t.criticality+(double)numCriticalityLevels+(double)((1L<<numCriticalityLevels)-overload)))/((double)numTasksPerCriticality[(int)t.criticality]));
//KARTHIK: OLD VALUE:				resiliency += (Math.pow(0.5,t.criticality))/((double)(1L<<numCriticalityLevels)*(double)numTasksPerCriticality[(int)t.criticality]);
				resiliency += (Math.pow(0.5,t.criticality))/((double)(2*numCriticalityLevels)*(double)numTasksPerCriticality[(int)t.criticality]);
				//System.out.println(t.name + " meets deadline Co=" + t.getOverloadedC() + " Cn=" + t.getNormalC() + "Crm = " + t.getCrm() + "T= " + t.getPeriod());
			}
			else
			{
				//System.out.println(t.name + " misses deadline Co=" + t.getOverloadedC() + " Cn=" + t.getNormalC() + "Crm = " + t.getCrm() + "T= " + t.getPeriod());
			}
		}
		//System.out.println("Returning " + resiliency);
		//System.out.println("Done with a processor ");
		return resiliency;
	}
}
