/*
Copyright (c) 2014 Carnegie Mellon University.
All Rights Reserved.
Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following acknowledgments and disclaimers.
2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following acknowledgments and disclaimers in the documentation and/or other materials provided with the distribution.
3. Products derived from this software may not include Carnegie Mellon University, "SEI and/or Software Engineering Institute" in the name of such derived product, nor shall Carnegie Mellon University, "SEI and/or Software Engineering Institute" be used to endorse or promote products derived from this software without prior written permission. For written permission, please contact permission@sei.cmu.edu.
ACKNOWLEDMENTS AND DISCLAIMERS:
Copyright 2014 Carnegie Mellon University

This material is based upon work funded and supported by the Department of Defense under Contract No. FA8721-05-C-0003 with Carnegie Mellon University for the operation of the Software Engineering Institute, a federally funded research and development center.

Any opinions, findings and conclusions or recommendations expressed in this material are those of the author(s) and do not necessarily reflect the views of the United States Department of Defense.

NO WARRANTY. THIS CARNEGIE MELLON UNIVERSITY AND SOFTWARE ENGINEERING INSTITUTE MATERIAL IS FURNISHED ON AN AS-IS BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER EXPRESSED OR IMPLIED, AS TO ANY MATTER INCLUDING, BUT NOT LIMITED TO, WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY, OR RESULTS OBTAINED FROM USE OF THE MATERIAL. CARNEGIE MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.

This material has been approved for public release and unlimited distribution.

Carnegie Mellon is registered in the U.S. Patent and Trademark Office by Carnegie Mellon University.

DM-0000890

 */

package edu.cmu.sei.mCOP;

import java.util.Comparator;
import java.util.HashMap;
import java.util.TreeSet;

public class DecreasingDominantCriticalityOverloadedTaskComparator implements Comparator<MTask> {

	public int compare(MTask t1, MTask t2) {
		if (t1.getUniqueID() == t2.getUniqueID())
			return 0;
		long t1DominantCriticality = t1.getSustainedCriticality();//t1.getDominantCriticality();
		long t2DominantCriticality = t2.getSustainedCriticality();//t2.getDominantCriticality();
		if (t1DominantCriticality == t2DominantCriticality){
			double t1OverloadedUtilization = t1.getOverloadedUtilization(t1.getDominantMode());
			double t2OverloadedUtilization = t2.getOverloadedUtilization(t2.getDominantMode());
			if (t1OverloadedUtilization == t2OverloadedUtilization)
				return (int) (t1.getUniqueID() - t2.getUniqueID());
			else if (t1OverloadedUtilization < t2OverloadedUtilization)
				return 1;
			else
				return -1;
		} else {
			if (t1DominantCriticality > t2DominantCriticality)
				return 1;
			else
				return -1;
		}
	}
	
	// Test
	public static void main(String args[]){
		 final Mode m1 = new Mode();
		 final Mode m2 = new Mode();
		 
//		 String name, long T, long Ca, long Cs, HashMap<Mode,Long> criticalities,
//			long RMPriority
		 MTask t1 = new MTask("t1",10, 10, 10, new HashMap<Mode,Long>(){{put(m1,0l);put(m2,1l);}},1,2);
		 MTask t2 = new MTask("t2",10,10,10, new HashMap<Mode,Long>(){{put(m1,1l);put(m2,2l);}},1,2);
		 MTask t3 = new MTask("t3",10, 20, 10, new HashMap<Mode,Long>(){{put(m1,0l);put(m2,1l);}},1,2);
		 MTask t4 = new MTask("t4",10,20,10, new HashMap<Mode,Long>(){{put(m1,1l);put(m2,2l);}},1,2);
		 
		 TreeSet<MTask> set = new TreeSet<MTask>(new DecreasingDominantCriticalityOverloadedTaskComparator());
		 set.add(t1);
		 set.add(t2);
		 set.add(t3);
		 set.add(t4);
		 
		 for (MTask t:set){
			 //System.out.println("Task["+t.name+"].dominantCriticality:"+t.getDominantCriticality()+" Overload Utilization: "+t.getOverloadedUtilization(t.getDominantMode()));
			 System.out.println("Task["+t.name+"].sustainedCriticality:"+t.getSustainedCriticality()+" Overload Utilization: "+t.getOverloadedUtilization(t.getDominantMode()));
		 }
	}
}
