/*
Copyright (c) 2014 Carnegie Mellon University.
All Rights Reserved.
Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following acknowledgments and disclaimers.
2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following acknowledgments and disclaimers in the documentation and/or other materials provided with the distribution.
3. Products derived from this software may not include Carnegie Mellon University, "SEI and/or Software Engineering Institute" in the name of such derived product, nor shall Carnegie Mellon University, "SEI and/or Software Engineering Institute" be used to endorse or promote products derived from this software without prior written permission. For written permission, please contact permission@sei.cmu.edu.
ACKNOWLEDMENTS AND DISCLAIMERS:
Copyright 2014 Carnegie Mellon University

This material is based upon work funded and supported by the Department of Defense under Contract No. FA8721-05-C-0003 with Carnegie Mellon University for the operation of the Software Engineering Institute, a federally funded research and development center.

Any opinions, findings and conclusions or recommendations expressed in this material are those of the author(s) and do not necessarily reflect the views of the United States Department of Defense.

NO WARRANTY. THIS CARNEGIE MELLON UNIVERSITY AND SOFTWARE ENGINEERING INSTITUTE MATERIAL IS FURNISHED ON AN AS-IS BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER EXPRESSED OR IMPLIED, AS TO ANY MATTER INCLUDING, BUT NOT LIMITED TO, WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY, OR RESULTS OBTAINED FROM USE OF THE MATERIAL. CARNEGIE MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.

This material has been approved for public release and unlimited distribution.

Carnegie Mellon is registered in the U.S. Patent and Trademark Office by Carnegie Mellon University.

DM-0000890

 */

package edu.cmu.sei.mCOP;

import java.util.HashMap;
import java.util.TreeSet;
import java.util.Vector;

import edu.cmu.sei.ZeroSlackBinPacker.DualModeProcessor;
import edu.cmu.sei.ZeroSlackBinPacker.IncreasingCriticalityComparator;
import edu.cmu.sei.ZeroSlackBinPacker.PTask;
import edu.cmu.sei.ZeroSlackRM.Task;
import edu.cmu.sei.ZeroSlackRM.ZSRMScheduler;

public class MProcessor extends DualModeProcessor {

	HashMap<Mode,Double> modalFullnessLevel = new HashMap<Mode,Double>();
	HashMap<Mode,Double> modalNormalFullnessLevel = new HashMap<Mode,Double>();

	Vector<Mode> modes;
	Vector<MTask> globalMTaskset;
	
	public Vector<MTask> getGlobalMTaskset(){
		return globalMTaskset;
	}

	HashMap<Mode,Vector<MTask>> globalMTasksPerMode = new HashMap<Mode,Vector<MTask>>();
	
	public Vector<MTask> getGlobalMTasksetForMode(Mode m){
		Vector<MTask>v = globalMTasksPerMode.get(m);
		if (v ==null){
			v = new Vector<MTask>();
			globalMTasksPerMode.put(m,v);
			for (MTask t:getGlobalMTaskset()){
				if (t.getModalFacade(m) != null){
					v.add(t);
				}
			}
		}
		return v;
	}
	
	public double getOverloadedFullnessLevel(Mode m){
		return modalFullnessLevel.get(m);
	}
	
	public double getNormalFullnessLevel(Mode m){
		return modalNormalFullnessLevel.get(m);
	}

	public MProcessor(String name, Vector<PTask> gts, Vector<MTask> gmts, Vector<Mode> modes) {
		super(name, gts);
		this.globalMTaskset = gmts;
		this.modes = modes;
		for (Mode m:modes){
			modalFullnessLevel.put(m, 0.0);
			modalNormalFullnessLevel.put(m,0.0);
		}
	}
	
	public void addTask(MTask t){
		super.addTask(t);		
		for (Mode m: t.getActiveModes()){
			modalFullnessLevel.put(m, modalFullnessLevel.get(m)+t.getOverloadedUtilization(m));
			modalNormalFullnessLevel.put(m, modalNormalFullnessLevel.get(m)+t.getNormalUtilization(m));
		}
	}

	public void removeTask(MTask t){
		super.removeTask(t);
		for (Mode m:t.getActiveModes()){
			modalFullnessLevel.put(m, modalFullnessLevel.get(m)-t.getOverloadedUtilization(m));
			modalNormalFullnessLevel.put(m, modalNormalFullnessLevel.get(m)-t.getNormalUtilization(m));			
		}
	}

	public boolean isSchedulable(){
		boolean schedulable = true;
		if (compressionMode){
			for (Mode m:modes){
				schedulable = schedulable && (modalFullnessLevel.get(m) <=1.0);
			}
			return schedulable;
		} else {
			// multimodal compress schedule
			HashMap<Mode,ZSRMScheduler> schedulers = new HashMap<Mode,ZSRMScheduler>();
			
			for (Mode m:modes){
				ZSRMScheduler scheduler =new ZSRMScheduler(); 
				schedulers.put(m, scheduler);
		
				for (PTask t:getTaskSet()){
					scheduler.addTask(((MTask)t).getModalFacade(m));
				}

				scheduler.compressSchedule();

				for (Task t:scheduler.getTasksByPriority()){
					schedulable &= ((t.ZeroSlackInstant >=0) &&(t.getCrm()>=t.getNormalC()));
				}
			}

			return schedulable;			
		}
	}

	HashMap<Mode,Vector<PTask>> modalFacadesPerMode = new HashMap<Mode,Vector<PTask>>();
	public  Vector<PTask> getModalTaskset(Mode m){
		Vector <PTask> ts = modalFacadesPerMode.get(m);
		if (ts == null){
			ts = new Vector<PTask>();
			modalFacadesPerMode.put(m, ts);
			for (PTask t:getTaskSet()){
				ts.add(((MTask)t).getModalFacade(m));
			}
		}
		return ts;
	}

	public boolean addIfSchedulable(MTask t){
		addTask(t);
		if (!isSchedulable()){
			removeTask(t);
			return false;
		}
		return true;
	}

	public TreeSet<PTask> getTasksByCriticalityForMode(Mode m){
		TreeSet<PTask> ts = new TreeSet<PTask>(new IncreasingCriticalityComparator());
		for (PTask pt:this.getTaskSet()){
			MTask mt=(MTask) pt;
			PTask facade = mt.getModalFacade(m);
			if (facade !=null){
				ts.add(facade);
			}
		}
		return ts;
	}
	
	public double getModeResiliency(long overload, int numCriticalityLevels, Mode m){
		double resiliency = 0;
		long[] numTasksPerCriticality = new long[numCriticalityLevels];
		TreeSet<PTask> tasksByCriticality = getTasksByCriticalityForMode(m);//new TreeSet<PTask>(new IncreasingCriticalityComparator());		
		//tasksByCriticality.addAll(taskset);
		
		for(int i=0; i<numCriticalityLevels; i++){
			numTasksPerCriticality[i] = 0;
		}
		
		for (MTask t:getGlobalMTasksetForMode(m)){
			if(t.getModalFacade(m).criticality >= numCriticalityLevels){
				System.err.println("ERROR IN ASSUMPTIONS - CRITICALITY LEVEL ("+t.getModalFacade(m).criticality+") IS GREATER THAN NUMCRITICALITYLEVELS");
				return 0;
			}
			numTasksPerCriticality[(int)t.getModalFacade(m).criticality] ++;
		}
		for (PTask t:tasksByCriticality){
			double preemption=0;
			
			for (PTask pt:getModalTaskset(m)){
				if(t.getUniqueID() == pt.getUniqueID())
					continue;
				// preemption from lower priority higher or same criticality
				if (pt.criticality <= t.criticality && pt.RmPriority > t.RmPriority){
					//check if it is active in our overload condition
					if((overload&(1L<<pt.criticality))>0)
					{
						//Active in our overload scenario
						preemption += ((double)(pt.getOverloadedC()-pt.getCrm())/(double)pt.getPeriod());
					}
				} else if (pt.RmPriority <= t.RmPriority && pt.criticality> t.criticality){
					//higher priority but lower criticality
					preemption += ((double)pt.getCrm()/(double)pt.getPeriod());
				} else if (pt.RmPriority <= t.RmPriority){
					//higher priority and higher criticality
					if((overload&(1L<<pt.criticality))>0)
					{
						preemption += ((double)pt.getOverloadedC()/(double) pt.getPeriod());
					}
					else
					{
						preemption += ((double)pt.getNormalC()/(double)pt.getPeriod());	
					}
				}				
			}
			
			if (t.getOverloadedUtilization()+preemption <=1.0){
				resiliency += (Math.pow(0.5,t.criticality))/((double)(2*numCriticalityLevels)*(double)numTasksPerCriticality[(int)t.criticality]);
			}
			else
			{
				//System.out.println(t.name + " misses deadline Co=" + t.getOverloadedC() + " Cn=" + t.getNormalC() + "Crm = " + t.getCrm() + "T= " + t.getPeriod());
			}
		}
		//System.out.println("Returning " + resiliency);
		//System.out.println("Done with a processor ");
		return resiliency;
	}

	// Testing
	public static void main(String args[]){
		final Mode m1 = new Mode();
		final Mode m2 = new Mode();
		
		Vector<Mode> modes = new Vector<Mode>();
		modes.add(m1);
		modes.add(m2);
		
		MTask t1 = new MTask("t1", 4, 2, 1,new HashMap<Mode,Long> (){{put(m1,2l);put(m2,1l);}},1,2);
		MTask t2 = new MTask("t2", 8, 5, 2,new HashMap<Mode,Long> (){{put(m1,1l);put(m2,2l);}},2,2);
		
		MProcessor p = new MProcessor("p",null, null, modes);
		System.out.println("p1.addIfSchedulable(t["+t1.name+"]):"+p.addIfSchedulable(t1));
		System.out.println("p1.addIfSchedulable(t["+t2.name+"]):"+p.addIfSchedulable(t2));
		
		System.out.println("t1.m1 zs: "+t1.getModalFacade(m1).ZeroSlackInstant);
		System.out.println("t1.m2 zs: "+t1.getModalFacade(m2).ZeroSlackInstant);
		System.out.println("t2.m1 zs: "+t2.getModalFacade(m1).ZeroSlackInstant);
		System.out.println("t2.m2 zs: "+t2.getModalFacade(m2).ZeroSlackInstant);
	}
}
