/*
Copyright (c) 2014 Carnegie Mellon University.
All Rights Reserved.
Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following acknowledgments and disclaimers.
2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following acknowledgments and disclaimers in the documentation and/or other materials provided with the distribution.
3. Products derived from this software may not include Carnegie Mellon University, "SEI and/or Software Engineering Institute" in the name of such derived product, nor shall Carnegie Mellon University, "SEI and/or Software Engineering Institute" be used to endorse or promote products derived from this software without prior written permission. For written permission, please contact permission@sei.cmu.edu.
ACKNOWLEDMENTS AND DISCLAIMERS:
Copyright 2014 Carnegie Mellon University

This material is based upon work funded and supported by the Department of Defense under Contract No. FA8721-05-C-0003 with Carnegie Mellon University for the operation of the Software Engineering Institute, a federally funded research and development center.

Any opinions, findings and conclusions or recommendations expressed in this material are those of the author(s) and do not necessarily reflect the views of the United States Department of Defense.

NO WARRANTY. THIS CARNEGIE MELLON UNIVERSITY AND SOFTWARE ENGINEERING INSTITUTE MATERIAL IS FURNISHED ON AN AS-IS BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER EXPRESSED OR IMPLIED, AS TO ANY MATTER INCLUDING, BUT NOT LIMITED TO, WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY, OR RESULTS OBTAINED FROM USE OF THE MATERIAL. CARNEGIE MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.

This material has been approved for public release and unlimited distribution.

Carnegie Mellon is registered in the U.S. Patent and Trademark Office by Carnegie Mellon University.

DM-0000890

 */

package edu.cmu.sei.mCOP;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import edu.cmu.sei.ZeroSlackBinPacker.PTask;

public class MTask extends PTask {	
	HashMap<Mode,Long> modeToCriticality = new HashMap<Mode,Long>();
	HashMap<Mode, PTask> modalFacade = new HashMap<Mode, PTask>();
	
	long sustainedCriticality=0;
	
	int numModes=0;
	
	public void setNumModes(int n){
		numModes = n;
	}
	
	public int getNumModes(){
		return numModes;
	}
	
	// the dominant criticality is the smallest. Smaller value means larger criticality
	long dominantCriticality= Long.MAX_VALUE;
	
	long transZeroSlackInstant;
	
	long lingeringTransZeroSlackInstant;
		
	public HashMap<Mode,Long> getModeCriticalities(){
		return modeToCriticality;
	}
	
	Mode dominantMode;
	
	public long getSustainedCriticality(){
		return sustainedCriticality;
	}
	
	public long getDominantCriticality(){
		return dominantCriticality;
	}
	
	public Mode getDominantMode(){
		return dominantMode;
	}
	
	public Set<Mode> getActiveModes(){
		return modeToCriticality.keySet();
	}
	
	public PTask getModalFacade(Mode m){
		return modalFacade.get(m);
	}
	
	public Collection<PTask> getAllActiveModalFacades(){
		return modalFacade.values();
	}
	
	public void addToSustainedCriticality(long crit){
		sustainedCriticality += (long)Math.pow((double)(numModes+1),(double)(numModes - crit));
	}
	
	public MTask(String name, long T, long Ca, long Cs, HashMap<Mode,Long> criticalities,
			long RMPriority, int nModes) {		
		super(name, T, Ca, Cs, criticalities.get(criticalities.keySet().iterator().next()), RMPriority);
		numModes = nModes;
		modeToCriticality = criticalities;

		long criticality;
		for (Map.Entry<Mode,Long> entry:modeToCriticality.entrySet()){
			criticality = entry.getValue();
			//sustainedCriticality += criticality;
			addToSustainedCriticality(criticality);
			modalFacade.put(entry.getKey(), new PTask(name,T,Ca,Cs,criticality, RMPriority));
			if(dominantCriticality > criticality){
				dominantCriticality = criticality;
				dominantMode = entry.getKey();
			}
		}
	}
	
	public void setPriority(int p){
		RmPriority = p;
		for (PTask t:this.modalFacade.values()){
			t.RmPriority = p;
		}
	}
	
	public double getOverloadedUtilization(Mode m){
		if (modeToCriticality.keySet().contains(m)){
			return getOverloadedUtilization();
		}
		return 0.0;
	}
	
	public double getNormalUtilization(Mode m){
		if (modeToCriticality.keySet().contains(m)){
			return getNormalUtilization();
		}
		return 0.0;
	}
	
	public static MTask cloneTask(MTask base){
		MTask dup = new MTask(base.name,
							base.getPeriod(),
							base.getOverloadedC(),
							base.getNormalC(),
							base.modeToCriticality,
							base.RmPriority, base.getNumModes());
		return dup;
	}
}
